package ezreader10;

import java.io.PrintStream;
import java.util.*;

class Stats {
    
    //**************************************************************************
     
    void analyses(PrintStream diskWriter, Display display, ArrayList<Sentence> text) {
        
        // Calculate mean word-based measures (across subjects) for every word in sentence corpus:
        means(text);
        
        // Calculate frequency-class means, RMSD, and mean regression probability:
        classMeans(diskWriter, text);
        
        // Display mean word-based measures and distributions for all words:
        if (EZReader10.displayMeans == true) {
            display.means(diskWriter, text);
            display.fixDists(diskWriter, text);
            display.refixProbs(diskWriter, text);
            display.IOVPs(diskWriter, text);
        }
           
        // Display mean distributions for words of each given length:
        if (EZReader10.displayDists == true) {
                        
            // Initialize classes for first-fixation landing-site distributions:
            ArrayList<Word> fixDists = new ArrayList<>();
            for (int i = 0; i < EZReader10.maxLength; i++) {
                Word fixDist = new Word();
                fixDists.add(fixDist);
            }
                    
            // Calculate & display first-fixation landing-site distributions:
            fixDists(text, fixDists);
            display.meanFixDists(diskWriter, fixDists);
            fixDists.clear();
                        
            // Initialize classes for refixation-probability distributions:
            ArrayList<Word> refixProbs = new ArrayList<>();
            for (int i = 0; i < EZReader10.maxLength; i++) {
                Word refixProb = new Word();
                refixProbs.add(refixProb);
            }
                    
            // Calculate & display refixation-probability distributions:
            refixProbs(text, refixProbs);
            display.meanRefixProbs(diskWriter, refixProbs); 
                                
            // Initialize classes for IOVPs:
            ArrayList<Word> IOVPs = new ArrayList<>();
            for (int i = 0; i < EZReader10.maxLength; i++) {
                Word IOVP = new Word();
                IOVPs.add(IOVP);
            }
                    
            // Calculate & display IOVPs:
            IOVPs(text, IOVPs);
            display.meanIOVPs(diskWriter, IOVPs);                       
        }
    }
    
    //**************************************************************************
     
    void classMeans(PrintStream diskWriter, ArrayList<Sentence> text) {
        
        Display display = new Display();
/*        
        // Counter to tally # words per frequency class:
        int[] N = new int[] { 0, 0, 0, 0, 0 };  
      
        // Initialize classes for frequency-class means:
        ArrayList<Word> classMeans = new ArrayList<>();
        for (int i = 0; i < EZReader10.NFreqClass; i++) {
            Word classMean = new Word();
            classMeans.add(classMean);
        }
        
        // Sum DVs for each class:
        for (int i = 0; i < EZReader10.NSentences; i++) {
            for (int j = 1; j < text.get(i).numberWords - 1; j++) {
                int freqClass = text.get(i).get(j).iv.freqClass;
                classMeans.get(freqClass).dv.SFD += text.get(i).get(j).dv.SFD;
                classMeans.get(freqClass).dv.FFD += text.get(i).get(j).dv.FFD;
                classMeans.get(freqClass).dv.GD += text.get(i).get(j).dv.GD;
                classMeans.get(freqClass).dv.Pr1 += text.get(i).get(j).dv.Pr1;
                classMeans.get(freqClass).dv.Pr2 += text.get(i).get(j).dv.Pr2;
                classMeans.get(freqClass).dv.PrS += text.get(i).get(j).dv.PrS;
                N[freqClass]++;
            }
        }
        
        // Calculate class means:
        for (int i = 0; i < EZReader10.NFreqClass; i++) {
            classMeans.get(i).dv.SFD /= N[i];
            classMeans.get(i).dv.FFD /= N[i];
            classMeans.get(i).dv.GD /= N[i];
            classMeans.get(i).dv.Pr1 /= N[i];
            classMeans.get(i).dv.Pr2 /= N[i];
            classMeans.get(i).dv.PrS /= N[i];
        }
        
        // Initialize observed means and standard deviations (based on Schilling et al., 1998):
        double[] obsMeanFFD = new double[] { 251, 239, 237, 233, 216 }; 
        double[] obsSDFFD = new double[] { 30.13, 36.83, 35.47, 56.60, 75.15 };
        double[] obsMeanSFD = new double[] { 264, 248, 243, 237, 216 };
        double[] obsSDSFD = new double[] { 36.20, 41.41, 37.14, 63.76, 75.09 };  
        double[] obsMeanGD = new double[] { 308, 278, 265, 241, 222 }; 
        double[] obsSDGD = new double[] { 56.40, 53.72, 51.88, 63.28, 77.10 }; 
        double[] obsMeanPr1 = new double[] { 0.67, 0.70, 0.68, 0.44, 0.31 }; 
        double[] obsSDPr1 = new double[] { 0.2196, 0.2100, 0.2169, 0.2463, 0.2157 }; 
        double[] obsMeanPr2 = new double[] { 0.23, 0.18, 0.10, 0.02, 0.01 };
        double[] obsSDPr2 = new double[] { 0.1747, 0.1450, 0.0916, 0.0188, 0.0108 };
        double[] obsMeanPrS = new double[] { 0.10, 0.12, 0.22, 0.54, 0.67 };  
        double[] obsSDPrS = new double[] { 0.0900, 0.1087, 0.1694, 0.2483, 0.2196 }; 
        
        // Calculate RMSD:
        double sum = 0;
        double RMSD = 0;
        for (int i = 0; i < EZReader10.NFreqClass; i++) {
            sum += Math.pow(((obsMeanFFD[i] - classMeans.get(i).dv.FFD) / obsSDFFD[i]), 2);
            sum += Math.pow(((obsMeanSFD[i] - classMeans.get(i).dv.SFD) / obsSDSFD[i]), 2);
            sum += Math.pow(((obsMeanGD[i] - classMeans.get(i).dv.GD) / obsSDGD[i]), 2);
            sum += Math.pow(((obsMeanPr1[i] - classMeans.get(i).dv.Pr1) / obsSDPr1[i]), 2);
            sum += Math.pow(((obsMeanPr2[i] - classMeans.get(i).dv.Pr2) / obsSDPr2[i]), 2);
            sum += Math.pow(((obsMeanPrS[i] - classMeans.get(i).dv.PrS) / obsSDPrS[i]), 2);
        }
        RMSD = Math.sqrt(sum / 30.0);
*/        
       // Counter to tally # words per frequency class:
        int[] N = new int[] { 0, 0, 0, 0, 0, 0, 0, 0 };  
      
        // Initialize classes for frequency-class means:
        ArrayList<Word> classMeans = new ArrayList<>();
        for (int i = 0; i < EZReader10.NFreqClass; i++) {
            Word classMean = new Word();
            classMeans.add(classMean);
        }
        
        // Sum DVs for each class:
        for (int i = 0; i < EZReader10.NSentences; i++) {
            for (int j = 1; j < text.get(i).numberWords - 1; j++) {
                int freqClass = text.get(i).get(j).iv.freqClass;
                classMeans.get(freqClass).dv.SFD += text.get(i).get(j).dv.SFD;
                classMeans.get(freqClass).dv.FFD += text.get(i).get(j).dv.FFD;
                classMeans.get(freqClass).dv.GD += text.get(i).get(j).dv.GD;
                classMeans.get(freqClass).dv.Pr1 += text.get(i).get(j).dv.Pr1;
                classMeans.get(freqClass).dv.Pr2 += text.get(i).get(j).dv.Pr2;
                classMeans.get(freqClass).dv.PrS += text.get(i).get(j).dv.PrS;
                N[freqClass]++;
            }
        }
        
        // Calculate class means:
        for (int i = 0; i < EZReader10.NFreqClass; i++) {
            classMeans.get(i).dv.SFD /= N[i];
            classMeans.get(i).dv.FFD /= N[i];
            classMeans.get(i).dv.GD /= N[i];
            classMeans.get(i).dv.Pr1 /= N[i];
            classMeans.get(i).dv.Pr2 /= N[i];
            classMeans.get(i).dv.PrS /= N[i];
        }
     
        // Initialize observed means and standard deviations (based on Schilling et al., 1998):
        double[] obsMeanFFD = new double[] { 274, 237, 243, 246, 238, 242, 224, 205 }; 
        double[] obsSDFFD = new double[] { 42, 20, 20, 37, 36, 37, 74, 84 };       
        double[] obsMeanSFD = new double[] { 281, 257, 253, 260, 246, 246, 228, 204 }; 
        double[] obsSDSFD = new double[] { 52, 21, 26, 45, 37, 39, 82, 84 };        
        double[] obsMeanGD = new double[] { 336, 284, 304, 301, 266, 260, 235, 209 }; 
        double[] obsSDGD = new double[] { 60, 26, 53, 65, 45, 52, 82, 86 };        
        double[] obsMeanPr1 = new double[] { 0.725, 0.655, 0.695, 0.669, 0.714, 0.575, 0.373, 0.311 }; 
        double[] obsSDPr1 = new double[] { 0.446, 0.475, 0.460, 0.471, 0.452, 0.494, 0.484, 0.463 }; // SD = p (1 - p) (i.e., as per Reichle et al., 1998)       
        double[] obsMeanPr2 = new double[] { 0.203, 0.177, 0.254, 0.223, 0.128, 0.053, 0.025, 0.008 }; 
        double[] obsSDPr2 = new double[] { 0.402, 0.382, 0.435, 0.416, 0.334, 0.224, 0.157, 0.088 };        
        double[] obsMeanPrS = new double[] { 0.071, 0.167, 0.051, 0.108, 0.158, 0.372, 0.601, 0.681 };  
        double[] obsSDPrS = new double[] { 0.258, 0.373, 0.220, 0.311, 0.364, 0.483, 0.490, 0.466 }; 
        
        // Calculate RMSD:
        double sum = 0;
        for (int i = 0; i < EZReader10.NFreqClass; i++) {
            sum += Math.pow(((obsMeanFFD[i] - classMeans.get(i).dv.FFD) / obsSDFFD[i]), 2);
            sum += Math.pow(((obsMeanSFD[i] - classMeans.get(i).dv.SFD) / obsSDSFD[i]), 2);
            sum += Math.pow(((obsMeanGD[i] - classMeans.get(i).dv.GD) / obsSDGD[i]), 2);
            sum += Math.pow(((obsMeanPr1[i] - classMeans.get(i).dv.Pr1) / obsSDPr1[i]), 2);
            sum += Math.pow(((obsMeanPr2[i] - classMeans.get(i).dv.Pr2) / obsSDPr2[i]), 2);
            sum += Math.pow(((obsMeanPrS[i] - classMeans.get(i).dv.PrS) / obsSDPrS[i]), 2);
        }
        EZReader10.RMSD = Math.sqrt(sum / 48.0); // Note: 48 = # means
        if (EZReader10.RMSD < EZReader10.bestRMSD) EZReader10.bestRMSD = EZReader10.RMSD;

        // Calculate mean probability of regression:
        double PrReg = 0;
        for (int i = 0; i < EZReader10.NSentences; i++) PrReg += text.get(i).regressionN;
        PrReg /= (EZReader10.NSentences * EZReader10.NSubjects);
        
        System.out.printf(" %5.4f %5.4f\n", EZReader10.RMSD, EZReader10.bestRMSD);
       
        if (EZReader10.displayClassMeans == true) display.classMeans(diskWriter, classMeans, EZReader10.RMSD, PrReg);
    }

    //**************************************************************************
    
    void fixDists(ArrayList<Sentence> text, ArrayList<Word> fixDists) {
        
        // Counter to tally # of words of each length:
        int N[] = new int[EZReader10.maxLength];
        for (int i = 0; i < EZReader10.maxLength; i++) N[i] = 0;
        
        // Calculate distributions:
        for (int length = 0; length < EZReader10.maxLength; length++) {
            
            for (int i = 0; i < EZReader10.NSentences; i++) {
                for (int j = 1; j < text.get(i).numberWords - 1; j++) {
                    if (text.get(i).get(j).iv.length == length) {
                        if (text.get(i).get(j).dv.Pr1 != 0) {
                            for (int k = 0; k < EZReader10.maxLength; k++) fixDists.get(length).dv.FixDist[k] += text.get(i).get(j).dv.FixDist[k];
                            N[length]++;
                        }
                    }
                }
            }
        }
        
        for (int length = 0; length < EZReader10.maxLength; length++) {
            for (int i = 0; i < EZReader10.maxLength; i++) if (N[length] > 0) fixDists.get(length).dv.FixDist[i] /= N[length];
        }
    }
    
    //**************************************************************************
     
    void IOVPs(ArrayList<Sentence> text, ArrayList<Word> IOVPs) {
        
        // Calculate distributions:
        for (int length = 0; length < EZReader10.maxLength; length++) {
            
            for (int i = 0; i < EZReader10.NSentences; i++) {
                for (int j = 1; j < text.get(i).numberWords - 1; j++) {
                    if (text.get(i).get(j).iv.length == length) {
                        for (int k = 0; k < EZReader10.maxLength; k++) {
                            if (text.get(i).get(j).dv.IOVP[k] > 0) {
                                IOVPs.get(length).dv.IOVP[k] += text.get(i).get(j).dv.IOVP[k]; 
                                IOVPs.get(length).dv.N[k]++;
                            }
                        }
                    }
                }
            }
        }
        
        for (int length = 0; length < EZReader10.maxLength; length++) {
            for (int i = 0; i < EZReader10.maxLength; i++) if (IOVPs.get(length).dv.N[i] > 0) IOVPs.get(length).dv.IOVP[i] /= IOVPs.get(length).dv.N[i];
        }
    }
    
    //**************************************************************************
     
    void means(ArrayList<Sentence> text) {
        
        // Fixation-duration measures:
        for (int i = 0; i < EZReader10.NSentences; i++) {
            for (int j = 1; j < text.get(i).numberWords - 1; j++) {
                
                if (text.get(i).get(j).dv.Pr1 > 0) text.get(i).get(j).dv.SFD /= text.get(i).get(j).dv.Pr1;
         
                if (text.get(i).get(j).dv.PrF > 0) text.get(i).get(j).dv.TT /= text.get(i).get(j).dv.PrF;
                
                if ((text.get(i).get(j).dv.Pr1 + text.get(i).get(j).dv.Pr2) > 0) {
                    text.get(i).get(j).dv.FFD /= (text.get(i).get(j).dv.Pr1 + text.get(i).get(j).dv.Pr2);
                    text.get(i).get(j).dv.GD /= (text.get(i).get(j).dv.Pr1 + text.get(i).get(j).dv.Pr2);

                    // Distributions:
                    for (int k = 0; k < EZReader10.maxLength; k++) {
                        if (text.get(i).get(j).dv.FixDist[k] > 0) text.get(i).get(j).dv.RefixProb[k] /= text.get(i).get(j).dv.FixDist[k];
                        text.get(i).get(j).dv.FixDist[k] /= (text.get(i).get(j).dv.Pr1 + text.get(i).get(j).dv.Pr2);                        
                        if (text.get(i).get(j).dv.N[k] > 0) text.get(i).get(j).dv.IOVP[k] /= text.get(i).get(j).dv.N[k];
                    }
                }
             
                // Fixation-prob. measures:
                if (EZReader10.includeRegressions == true) text.get(i).regressionN = 0; 
                text.get(i).get(j).dv.Pr1 /= (EZReader10.NSubjects - text.get(i).regressionN);
                text.get(i).get(j).dv.Pr2 /= (EZReader10.NSubjects - text.get(i).regressionN);
                text.get(i).get(j).dv.PrS /= (EZReader10.NSubjects - text.get(i).regressionN);
                text.get(i).get(j).dv.PrF /= (EZReader10.NSubjects - text.get(i).regressionN);
            }
        }
    }

    //**************************************************************************
     
    void refixProbs(ArrayList<Sentence> text, ArrayList<Word> refixProbs) {
        
        // Counter to tally # of words of each length:
        int N[] = new int[EZReader10.maxLength];
        for (int i = 0; i < EZReader10.maxLength; i++) N[i] = 0;
        
        // Caclulate distributions:
        for (int length = 0; length < EZReader10.maxLength; length++) {
            for (int i = 0; i < EZReader10.NSentences; i++) {
                for (int j = 1; j < text.get(i).numberWords - 1; j++) {
                    if (text.get(i).get(j).iv.length == length) {
                        for (int k = 0; k < EZReader10.maxLength; k++) refixProbs.get(length).dv.RefixProb[k] += text.get(i).get(j).dv.RefixProb[k];
                        N[length]++;
                    }
                }
            }
        }
        
        for (int length = 0; length < EZReader10.maxLength; length++) {
            for (int i = 0; i < EZReader10.maxLength; i++) if (N[length] > 0) refixProbs.get(length).dv.RefixProb[i] /= N[length];
        }
    }
    
    //**************************************************************************
     
    void wordDVs(ArrayList<Sentence> text, int S, ArrayList<Fixation> trace) {
        
        for (int i = 1; i < text.get(S).numberWords - 1; i++) {
            
            // Initialize word class to hold DVs:
            Word w = new Word();
      
            // Convert fixation trace into word DVs:
            for (int j = 0; j < trace.size(); j++) {
                if (trace.get(j).word == i) {
                    // Identify possible interword regression:
                    boolean regression = false;
                    for (int k = 0; k <= j; k++) if (trace.get(k).word > i) regression = true;
                    
                    // If saccade was regression, increment NFix and TT:
                    if (regression == true) {
                        w.dv.NFix++;
                        w.dv.TT += trace.get(j).dur;
                    }
                    // If saccade was progressive:
                    else {
                        w.dv.NFix++; // increment # fixations
                        
                        // If fixation is first:
                        if (w.dv.FFD == 0) {
                            w.dv.FFD = trace.get(j).dur; // increment FFD
                            w.dv.Pos1 = (int)(trace.get(j).pos - text.get(S).get(trace.get(j).word).iv.pos0); // calculate within-word fixation location
                            w.dv.FixDist[w.dv.Pos1]++; // record fixation location for fixation landing-site distribution
                            if (j < (trace.size() - 1) && i == trace.get(j + 1).word) w.dv.RefixProb[w.dv.Pos1]++; // record fixation location for refixation-probability distribution
                            w.dv.GD = trace.get(j).dur; // increment GD
                        } // If fixation is not the first:
                        else if (j == 0 || (j > 0 && trace.get(j - 1).word == i)) w.dv.GD += trace.get(j).dur; // increment GD
                        w.dv.TT += trace.get(j).dur; // increment TT
                    }
                }
            }
            
            // Calculate DVs derived indirectly from trace:
            if (w.dv.FFD == 0) w.dv.PrS++; // increment # skips
            else {
                //w.dv.PrF++; // increment # fixations
                if (w.dv.FFD == w.dv.GD) { // i.e., word was fixated only once during first pass
                    w.dv.SFD = w.dv.FFD; // SFD = FFD
                    w.dv.Pr1++; // increment # single-fixation durations
                    // Increment arrays for IOVPs:
                    w.dv.IOVP[w.dv.Pos1] = w.dv.SFD; // durations
                    w.dv.N[w.dv.Pos1]++; // increment IOVP fixation-location count
                }
                else w.dv.Pr2++; // increment # multiple fixations
            }
            if (w.dv.TT > 0) w.dv.PrF++;
            
            // Add values to running totals used for means across subjects:
            text.get(S).get(i).dv.SFD += w.dv.SFD;
            text.get(S).get(i).dv.FFD += w.dv.FFD;
            text.get(S).get(i).dv.GD += w.dv.GD;
            text.get(S).get(i).dv.TT += w.dv.TT;
            text.get(S).get(i).dv.Pr1 += w.dv.Pr1;
            text.get(S).get(i).dv.Pr2 += w.dv.Pr2;
            text.get(S).get(i).dv.PrS += w.dv.PrS;
            text.get(S).get(i).dv.PrF += w.dv.PrF;
            text.get(S).get(i).dv.NFix += w.dv.NFix;
            for (int j = 0; j < EZReader10.maxLength; j++) {
                text.get(S).get(i).dv.FixDist[j] += w.dv.FixDist[j];
                text.get(S).get(i).dv.RefixProb[j] += w.dv.RefixProb[j];
                text.get(S).get(i).dv.IOVP[j] += w.dv.IOVP[j];
                text.get(S).get(i).dv.N[j] += w.dv.N[j];
            }
        } 
    }
}

//******************************************************************************

